"""Metadata module for managing Object Access Controls."""
# this will be autogenerated by pop-create with the resource id based on parameterOrder value from JSON schema for GET
RESOURCE_IDS = ["bucket", "object", "entity"]
